package fr.mrdev023.vulkan_java.vk;

import org.lwjgl.PointerBuffer;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.*;
import org.tinylog.Logger;

import fr.mrdev023.vulkan_java.vk.utils.SuitablePhysicalDeviceFinder;

import java.nio.*;
import java.util.*;

import static fr.mrdev023.vulkan_java.vk.VulkanError.vkCheck;
import static org.lwjgl.vulkan.VK11.*;

public class Device {

    private final PhysicalDevice physicalDevice;
    private final VkDevice vkDevice;
    private final Queue.GraphicsQueue graphicsQueue;
    private final Queue.ComputeQueue computeQueue;
    private final Queue.TransferQueue transferQueue;

    public Device(SuitablePhysicalDeviceFinder.MatchResult physicalDeviceMatch) throws VulkanError {
        this.physicalDevice = physicalDeviceMatch.physicalDevice;

        try (MemoryStack stack = MemoryStack.stackPush()) {
            // Define required extensions
            Set<String> deviceExtensions = getDeviceExtensions();
            boolean usePortability = deviceExtensions
                    .contains(KHRPortabilitySubset.VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME)
                    && VulkanUtils.getOS() == VulkanUtils.OSType.MACOS;

            int numExtensions = usePortability ? 2 : 1;
            PointerBuffer requiredExtensions = stack.mallocPointer(numExtensions);
            requiredExtensions.put(stack.ASCII(KHRSwapchain.VK_KHR_SWAPCHAIN_EXTENSION_NAME));
            if (usePortability) {
                requiredExtensions.put(stack.ASCII(KHRPortabilitySubset.VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME));
            }
            requiredExtensions.flip();

            // Set up required features
            VkPhysicalDeviceFeatures features = VkPhysicalDeviceFeatures.calloc(stack);

            // Enable all the queue families
            VkQueueFamilyProperties.Buffer queuePropsBuff = physicalDevice.getVkQueueFamilyProps();
            int numQueuesFamilies = queuePropsBuff.capacity();
            VkDeviceQueueCreateInfo.Buffer queueCreationInfoBuf = VkDeviceQueueCreateInfo.calloc(numQueuesFamilies,
                    stack);
            for (int i = 0; i < numQueuesFamilies; i++) {
                FloatBuffer priorities = stack.callocFloat(queuePropsBuff.get(i).queueCount());
                queueCreationInfoBuf.get(i)
                        .sType(VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO)
                        .queueFamilyIndex(i)
                        .pQueuePriorities(priorities);
            }

            VkDeviceCreateInfo deviceCreateInfo = VkDeviceCreateInfo.calloc(stack)
                    .sType(VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO)
                    .ppEnabledExtensionNames(requiredExtensions)
                    .pEnabledFeatures(features)
                    .pQueueCreateInfos(queueCreationInfoBuf);

            PointerBuffer pp = stack.mallocPointer(1);
            vkCheck(vkCreateDevice(physicalDevice.getVkPhysicalDevice(), deviceCreateInfo, null, pp),
                    "Failed to create device");
            vkDevice = new VkDevice(pp.get(0), physicalDevice.getVkPhysicalDevice(), deviceCreateInfo);

            graphicsQueue = new Queue.GraphicsQueue(this, physicalDeviceMatch.graphicsQueueFamilyIndex, 0);
            computeQueue = new Queue.ComputeQueue(this, physicalDeviceMatch.computeQueueFamilyIndex, 1);
            transferQueue = new Queue.TransferQueue(this, physicalDeviceMatch.transferQueueFamilyIndex, 2);
        }

        Logger.debug("Vulkan device created");
    }

    public void destroy() {
        Logger.debug("Destroying Vulkan device");
        vkDestroyDevice(vkDevice, null);
    }

    private Set<String> getDeviceExtensions() throws VulkanError {
        Set<String> deviceExtensions = new HashSet<>();
        try (MemoryStack stack = MemoryStack.stackPush()) {
            // Get the number of device extensions
            IntBuffer numExtensionsBuf = stack.callocInt(1);
            int result = vkEnumerateDeviceExtensionProperties(physicalDevice.getVkPhysicalDevice(), (String) null,
                    numExtensionsBuf,
                    null);
            vkCheck(result, "Failed to enumerate device extension properties");

            int numExtensions = numExtensionsBuf.get(0);
            Logger.debug("Vulkan device supported extensions ({}):", numExtensions);

            // Get the device extensions
            VkExtensionProperties.Buffer propsBuff = VkExtensionProperties.calloc(numExtensions, stack);
            result = vkEnumerateDeviceExtensionProperties(physicalDevice.getVkPhysicalDevice(), (String) null,
                    numExtensionsBuf,
                    propsBuff);
            vkCheck(result, "Failed to enumerate device extension properties");

            for (int i = 0; i < numExtensions; i++) {
                VkExtensionProperties props = propsBuff.get(i);
                String extensionName = props.extensionNameString();
                deviceExtensions.add(extensionName);
                Logger.debug("\t - {}", extensionName);
            }
        }
        return deviceExtensions;
    }

    public PhysicalDevice getPhysicalDevice() {
        return physicalDevice;
    }

    public VkDevice getVkDevice() {
        return vkDevice;
    }

    public void waitIdle() {
        vkDeviceWaitIdle(vkDevice);
    }

    public Queue.GraphicsQueue getGraphicsQueue() {
        return graphicsQueue;
    }

    public Queue.ComputeQueue getComputeQueue() {
        return computeQueue;
    }

    public Queue.TransferQueue getTransferQueue() {
        return transferQueue;
    }
}