package fr.mrdev023.vulkan_java.vk.utils;

import java.nio.IntBuffer;
import java.util.HashSet;
import java.util.Set;

import org.lwjgl.PointerBuffer;
import org.lwjgl.glfw.GLFWVulkan;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.EXTDebugUtils;
import org.lwjgl.vulkan.KHRPortabilityEnumeration;
import org.lwjgl.vulkan.VK10;
import org.lwjgl.vulkan.VkExtensionProperties;
import org.lwjgl.vulkan.VkInstanceCreateInfo;
import org.tinylog.Logger;

import fr.mrdev023.vulkan_java.vk.VulkanUtils;

/**
 * This class is used to store the instance extensions to use to create the
 * Vulkan instance.
 * 
 * @see InstanceExtensions.Selector to select the instance extensions
 */
public class InstanceExtensions {
    private Set<String> instanceExtensions;

    private InstanceExtensions(Set<String> instanceExtensions) {
        this.instanceExtensions = instanceExtensions;
    }

    public boolean hasInstanceExtensions() {
        return !instanceExtensions.isEmpty();
    }

    public Set<String> getInstanceExtensions() {
        return instanceExtensions;
    }

    public PointerBuffer writeToStack(MemoryStack stack) {
        int numExtensions = instanceExtensions.size();

        PointerBuffer requiredExtensions = stack.mallocPointer(numExtensions);
        for (String extension : instanceExtensions) {
            requiredExtensions.put(stack.UTF8(extension));
        }
        requiredExtensions.flip();

        return requiredExtensions;
    }

    public void writeInstanceCreateInfoFlags(VkInstanceCreateInfo instanceInfo) {
        if (instanceExtensions.contains(KHRPortabilityEnumeration.VK_KHR_PORTABILITY_ENUMERATION_EXTENSION_NAME)) {
            instanceInfo.flags(KHRPortabilityEnumeration.VK_INSTANCE_CREATE_ENUMERATE_PORTABILITY_BIT_KHR);
        }
    }

    /**
     * This class is used to automatically select the instance extensions
     * 
     * <h5>Specification</h5>
     * <ul>
     * <li>Get the GLFW required extensions to create the Window Surface</li>
     * <li>Get the portability extensions if the OS needs it</li>
     * <li>Add Debug Utils extension if validation layers are enabled</li>
     * </ul>
     * 
     * <h5>See Also</h5>
     * <ul>
     * <li>{@link #selectInstanceExtensions()}</li>
     * <li>{@link #withValidationLayers(InstanceValidationLayers)}</li>
     * </ul>
     */
    public static class Selector {
        private InstanceValidationLayers validationLayers;

        public Selector withValidationLayers(InstanceValidationLayers validationLayers) {
            this.validationLayers = validationLayers;
            return this;
        }

        /**
         * Selects the instance extensions.
         * 
         * @return the set of instance extensions
         */
        public InstanceExtensions selectInstanceExtensions() {
            Set<String> instanceExtensions = getInstanceExtensions();
            log("Supported instance extensions", instanceExtensions);

            Set<String> glfwExtensions = getGLFWRequiredExtensions();
            Set<String> portabilityExtensions = getPortabilityExtensions(instanceExtensions);

            Set<String> selectedExtensions = new HashSet<>();
            selectedExtensions.addAll(glfwExtensions);
            selectedExtensions.addAll(portabilityExtensions);

            if (validationLayers != null && validationLayers.hasValidationLayers()) {
                selectedExtensions.add(EXTDebugUtils.VK_EXT_DEBUG_UTILS_EXTENSION_NAME);
            }
            log("Selected instance extensions", selectedExtensions);

            return new InstanceExtensions(selectedExtensions);
        }

        private Set<String> getInstanceExtensions() {
            Set<String> instanceExtensions = new HashSet<>();

            try (MemoryStack stack = MemoryStack.stackPush()) {
                IntBuffer numExtensionsBuf = stack.callocInt(1);
                VK10.vkEnumerateInstanceExtensionProperties((String) null, numExtensionsBuf, null);
                int numExtensions = numExtensionsBuf.get(0);

                VkExtensionProperties.Buffer instanceExtensionsProps = VkExtensionProperties.calloc(numExtensions,
                        stack);
                VK10.vkEnumerateInstanceExtensionProperties((String) null, numExtensionsBuf, instanceExtensionsProps);
                for (int i = 0; i < numExtensions; i++) {
                    VkExtensionProperties props = instanceExtensionsProps.get(i);
                    String extensionName = props.extensionNameString();
                    instanceExtensions.add(extensionName);
                }
            }

            return instanceExtensions;
        }

        private Set<String> getPortabilityExtensions(Set<String> instanceExtensions) {
            Set<String> portabilityExtensions = new HashSet<>();

            var osType = VulkanUtils.getOS();
            if (osType == VulkanUtils.OSType.MACOS) {
                if (!instanceExtensions
                        .contains(KHRPortabilityEnumeration.VK_KHR_PORTABILITY_ENUMERATION_EXTENSION_NAME)) {
                    throw new RuntimeException(
                            "Vulkan instance does not support portability enumeration extension but it's required for MacOS");
                }
                portabilityExtensions.add(KHRPortabilityEnumeration.VK_KHR_PORTABILITY_ENUMERATION_EXTENSION_NAME);
            }

            return portabilityExtensions;
        }

        private Set<String> getGLFWRequiredExtensions() {
            PointerBuffer glfwExtensionsBuffer = GLFWVulkan.glfwGetRequiredInstanceExtensions();
            if (glfwExtensionsBuffer == null) {
                throw new RuntimeException("Failed to find the GLFW platform surface extensions");
            }

            Set<String> glfwExtensions = new HashSet<>();
            for (int i = 0; i < glfwExtensionsBuffer.remaining(); i++) {
                glfwExtensions.add(glfwExtensionsBuffer.getStringUTF8(i));
            }
            return glfwExtensions;
        }

        private void log(String title, Set<String> layers) {
            Logger.debug("{} ({})", title, layers.size());
            for (String layer : layers) {
                Logger.debug(" - {}", layer);
            }
        }
    }
}
