package fr.mrdev023.vulkan_java.vk.utils;

import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.KHRSurface;
import org.lwjgl.vulkan.VK10;
import org.tinylog.Logger;

import fr.mrdev023.vulkan_java.vk.Instance;
import fr.mrdev023.vulkan_java.vk.PhysicalDevice;
import fr.mrdev023.vulkan_java.vk.Surface;
import fr.mrdev023.vulkan_java.vk.VulkanError;

public class SuitablePhysicalDeviceFinder {

    public static MatchResult findBestPhysicalDevice(Instance instance, Criteria criteria) throws VulkanError {
        List<MatchResult> matchedPhysicalDevices = new ArrayList<>();

        try (MemoryStack stack = MemoryStack.stackPush()) {
            List<PhysicalDevice> physicalDevices = PhysicalDevice.getPhysicalDevices(instance, stack);
            for (PhysicalDevice physicalDevice : physicalDevices) {
                var matchResult = checkPhysicalDevice(stack, physicalDevice, criteria);
                if (matchResult.isSuitable(criteria)) {
                    matchedPhysicalDevices.add(matchResult);
                }
            }
        }

        return matchedPhysicalDevices.stream().min(Comparator.comparingInt(MatchResult::getScore)).orElse(null);
    }

    private static MatchResult checkPhysicalDevice(MemoryStack stack, PhysicalDevice physicalDevice,
            Criteria criteria) {
        int graphicsQueueFamilyIndex = -1;
        int computeQueueFamilyIndex = -1;
        int transferQueueFamilyIndex = -1;
        boolean surfaceSupport = false;

        IntBuffer presentSupport = stack.ints(VK10.VK_FALSE);
        var vkQueueFamilyProps = physicalDevice.getVkQueueFamilyProps();
        for (int i = 0; i < vkQueueFamilyProps.capacity(); i++) {
            var vkQueueFamilyProp = vkQueueFamilyProps.get(i);

            if (graphicsQueueFamilyIndex == -1 && (vkQueueFamilyProp.queueFlags() & VK10.VK_QUEUE_GRAPHICS_BIT) != 0) {
                graphicsQueueFamilyIndex = i;
            }

            if (computeQueueFamilyIndex == -1 && (vkQueueFamilyProp.queueFlags() & VK10.VK_QUEUE_COMPUTE_BIT) != 0) {
                computeQueueFamilyIndex = i;
            }

            if (transferQueueFamilyIndex == -1 && (vkQueueFamilyProp.queueFlags() & VK10.VK_QUEUE_TRANSFER_BIT) != 0) {
                transferQueueFamilyIndex = i;
            }

            KHRSurface.vkGetPhysicalDeviceSurfaceSupportKHR(physicalDevice.getVkPhysicalDevice(), i,
                    criteria.withSurfaceSupport.getVkSurface(), presentSupport);
            if (presentSupport.get(0) == VK10.VK_TRUE) {
                surfaceSupport = true;
            }

            if ((!criteria.withGraphicsQueue || graphicsQueueFamilyIndex != -1) &&
                    (!criteria.withComputeQueue || computeQueueFamilyIndex != -1) &&
                    (!criteria.withTransferQueue || transferQueueFamilyIndex != -1) &&
                    (criteria.withSurfaceSupport == null || surfaceSupport)) {
                // We found a suitable queue family, we can break the loop
                break;
            }
        }

        return new MatchResult(physicalDevice, graphicsQueueFamilyIndex, computeQueueFamilyIndex,
                transferQueueFamilyIndex, surfaceSupport);
    }

    public static class MatchResult {
        public final PhysicalDevice physicalDevice;
        public final boolean surfaceSupport;
        public final int graphicsQueueFamilyIndex;
        public final int computeQueueFamilyIndex;
        public final int transferQueueFamilyIndex;

        public MatchResult(PhysicalDevice physicalDevice, int graphicsQueueFamilyIndex, int computeQueueFamilyIndex,
                int transferQueueFamilyIndex, boolean surfaceSupport) {
            this.physicalDevice = physicalDevice;
            this.graphicsQueueFamilyIndex = graphicsQueueFamilyIndex;
            this.computeQueueFamilyIndex = computeQueueFamilyIndex;
            this.transferQueueFamilyIndex = transferQueueFamilyIndex;
            this.surfaceSupport = surfaceSupport;
        }

        public boolean isSuitable(Criteria criteria) {
            Logger.debug("Checking physical device");
            Logger.debug("\tProperties");
            Logger.debug("\t\tName: {}", physicalDevice.getVkPhysicalDeviceProperties().deviceNameString());
            int apiVersion = physicalDevice.getVkPhysicalDeviceProperties().apiVersion();
            Logger.debug("\t\tAPI version: {}.{}.{}",
                    VK10.VK_API_VERSION_MAJOR(apiVersion),
                    VK10.VK_API_VERSION_MINOR(apiVersion),
                    VK10.VK_API_VERSION_PATCH(apiVersion));
            Logger.debug("\t\tDevice type: {}", physicalDevice.getVkPhysicalDeviceProperties().deviceType());
            Logger.debug("\tRequired supports checking report");

            if (!criteria.withGraphicsQueue) {
                Logger.debug("\t\t[SKIPPED] Graphics queue is not required");
            } else if (graphicsQueueFamilyIndex == -1) {
                Logger.debug("\t\t[FAILED] Graphics queue is not supported");
                return false;
            } else {
                Logger.debug("\t\t[OK] Graphics queue is supported (family index: {})", graphicsQueueFamilyIndex);
            }

            if (!criteria.withComputeQueue) {
                Logger.debug("\t\t[SKIPPED] Compute queue is not required");
            } else if (computeQueueFamilyIndex == -1) {
                Logger.debug("\t\t[FAILED] Compute queue is not supported");
                return false;
            } else {
                Logger.debug("\t\t[OK] Compute queue is supported (family index: {})", computeQueueFamilyIndex);
            }

            if (!criteria.withTransferQueue) {
                Logger.debug("\t\t[SKIPPED] Transfer queue is not required");
            } else if (transferQueueFamilyIndex == -1) {
                Logger.debug("\t\t[FAILED] Transfer queue is not supported");
                return false;
            } else {
                Logger.debug("\t\t[OK] Transfer queue is supported (family index: {})", transferQueueFamilyIndex);
            }

            if (criteria.withSurfaceSupport == null) {
                Logger.debug("\t\t[SKIPPED] Surface support is not required");
            } else if (!surfaceSupport) {
                Logger.debug("\t\t[FAILED] Surface support is not supported");
                return false;
            } else {
                Logger.debug("\t\t[OK] Surface support is supported");
            }

            if (criteria.extensions == null) {
                Logger.debug("\t\t[SKIPPED] Required extensions is empty");
            } else if (physicalDevice.getVkDeviceExtensions().stream()
                    .allMatch(extension -> criteria.extensions.contains(extension.extensionNameString()))) {
                Logger.debug("\t\t[FAILED] Required extensions are not supported [{}]",
                        String.join(", ", criteria.extensions));
                return false;
            } else {
                Logger.debug("\t\t[OK] Required extensions are supported [{}]", String.join(", ", criteria.extensions));
            }

            return true;
        }

        public int getScore() {
            int deviceType = physicalDevice.getVkPhysicalDeviceProperties().deviceType();

            return switch (deviceType) {
                case VK10.VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU -> 0;
                case VK10.VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU -> 1;
                case VK10.VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU -> 2;
                case VK10.VK_PHYSICAL_DEVICE_TYPE_CPU -> 3;
                case VK10.VK_PHYSICAL_DEVICE_TYPE_OTHER -> 4;
                default -> 5;
            };
        }
    }

    public static class Criteria {
        private boolean withGraphicsQueue = false;
        private boolean withComputeQueue = false;
        private boolean withTransferQueue = false;
        private Surface withSurfaceSupport = null;
        private Set<String> extensions = new HashSet<>();

        public Criteria withGraphicsQueue(boolean withGraphicsQueue) {
            this.withGraphicsQueue = withGraphicsQueue;
            return this;
        }

        public Criteria withSurfaceSupport(Surface withSurfaceSupport) {
            this.withSurfaceSupport = withSurfaceSupport;
            return this;
        }

        public Criteria withComputeQueue(boolean withComputeQueue) {
            this.withComputeQueue = withComputeQueue;
            return this;
        }

        public Criteria withTransferQueue(boolean withTransferQueue) {
            this.withTransferQueue = withTransferQueue;
            return this;
        }

        public Criteria withExtensions(Set<String> extensions) {
            this.extensions = extensions;
            return this;
        }
    }
}
